#!/usr/local/bin/python
"""
    apply homographies calculated with Visual SFM on windows to 
    gantry acquired lightfield images
    
    produces warped images, aligned

    http://wiki.roberttwomey.com/Lightfield

    Robert Twomey 2015
    rtwomey@uw.edu

"""

import os
import sys
import cv2
import math
import numpy as np
from numpy import linalg
from multiprocessing import Queue
import multiprocessing
import argparse
import pickle
from pylab import * 


def readNVMfile(nvmfile):
    infile = open(nvmfile, 'r')
    
    lines = infile.readlines()
    print "\n"

    numcams = int(lines[2])
    
    nvmdict = {}
    
    camcenters = []
    xcents = []
    ycents = []
    for camresult in lines[3:3+numcams]:
        filename, results = camresult.split('\t')
        filename = filename.split("\\")[-1]
        results = np.array(results.split(' '))
        #.astype(np.float)
        focallength = results[0].astype(np.float)
        quaternion = results[1:5].astype(np.float)
        camcenter = results[5:8].astype(np.float)
        radialdistortion = results[8].astype(np.float)
        nvmdict[filename] = (focallength, quaternion, camcenter)
        # xcents.append(camcenter[0])
        # ycents.append(camcenter[2])
        # print quaternion, camcenter
        
    # print nvmdict
    #print camcenters
    
    # plot to screen
    #scatter(xcents, ycents)
    #show()
    
    return nvmdict
    
def nvmwarp(base_img, next_img, f, basequat, basecent, nextquat, nextcent):
 
    global warpedpath
    
    # base_img_rgb = cv2.imread(base_img)
    next_img_rgb = cv2.imread(next_img)
    base_img_rgb = cv2.imread(base_img)
    
    img_w = int(base_img_rgb.shape[1] * 1.5)
    img_h = int(base_img_rgb.shape[0] * 1.5)
    
    w = base_img_rgb.shape[1]
    h = base_img_rgb.shape[0]
    
    # subtract base rotation/translation from curr camera
    cconj_basequat = basequat * [1, -1, -1, -1]
    q = nextquat * cconj_basequat
    t = nextcent #- basecent

    # print f
    # print basequat
    # print cconj_basequat
    # print nextquat
    # print q
    # print basecent
    # print nextcent
    print t

    # from bundler software DataInterface.h
    qq = sqrt(q[0]*q[0]+q[1]*q[1]+q[2]*q[2]+q[3]*q[3])
    if qq>0:
        qw=q[0]/qq
        qx=q[1]/qq
        qy=q[2]/qq
        qz=q[3]/qq
    else:
        qw = 1
        qx = qy = qz = 0

    m = np.zeros((3,3))
    
    m[0][0]=qw*qw + qx*qx- qz*qz- qy*qy
    m[0][1]=2*qx*qy -2*qz*qw
    m[0][2]=2*qy*qw + 2*qz*qx
    m[1][0]=2*qx*qy+ 2*qw*qz
    m[1][1]=qy*qy+ qw*qw - qz*qz- qx*qx
    m[1][2]=2*qz*qy- 2*qx*qw
    m[2][0]=2*qx*qz- 2*qy*qw
    m[2][1]=2*qy*qz + 2*qw*qx
    m[2][2]=qz*qz+ qw*qw- qy*qy- qx*qx
    
    print m
    
    # camera center
    c =  np.zeros((3, 1))
    for j in range(0, 3):
        c[j] = -(m[0][j]*t[0] + m[1][j]*t[1] + m[2][j]*t[2])
    
    print c
    
    
    # http://stackoverflow.com/questions/14387867/problems-with-a-perspective-homography
    # http://stackoverflow.com/questions/12288473/perspective-warping-in-opencv-based-on-know-camera-orientation
    
    R = m
    # R = np.identity(3 )
    
    x = 0
    y = 0
    z = 0
    f = 1
    
    x1 = (x-90) * (pi / 180)
    y1 = (y-90) * (pi / 180)
    z1 = (z-90) * (pi / 180)
    
    # print x1, y1, z1
    # # Create a rotation matrix
    # R_vec = array([[x1], [y1], [z1]])
    # R, jacobian = cv2.Rodrigues(R_vec)
    #
    # #Create trans mat and combine with translation matrix
    # Trans_Mat = array([[[1], [0], [-w/2]],
    #                     [[0], [1], [-h/2]],
    #                     [[0], [0], [1]]])
    #
    # R_T_Mat = dot(R, Trans_Mat)
    # print R_T_Mat
    #
    # R_T_Mat[2][2] += h
    #
    # #Create and combine with camera matrix
    # Intrinsic_Mat = array([[[h], [0], [w/2]],
    #                        [[0], [h], [h/2]],
    #                        [[0], [0], [1]]])
    #
    # # print Intrinsic_Mat.shape
    # # print R_T_Mat.shape
    # H = Intrinsic_Mat * R_T_Mat
    # print H



    # http://stackoverflow.com/questions/7019407/translating-and-rotating-an-image-in-3d-using-opencv

    alpha = x1
    dist = 0#100
    
    # Projection 2D -> 3D matrix
    # A1 = np.float32([[1], [0], [-w/2]],
    #             [[0], [1], [-h/2]],
    #             [[0], [0], [0]],
    #             [[0], [0], [1]]
    #     ])
    A1 = np.float32([[1, 0, -w/2],
                [0, 1, -h/2],
                [0, 0, 0],
                [0, 0, 1]
        ])

    # Rotation matrices around the X axis

    # R = np.float32([[[1], [0], [0], [0]],
    #     [[0], [cos(alpha)], [-sin(alpha)], [0]],
    #     [[0], [sin(alpha)], [cos(alpha)], [0]],
    #     [[0], [0], [0], [1]]])

    R = np.float32([[1, 0, 0, 0],
        [0, cos(alpha), -sin(alpha), 0],
        [0, sin(alpha), cos(alpha), 0],
        [0, 0, 0, 1]])        

    # Translation matrix on the Z axis
    T = np.float32([[1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 1, dist],
        [0, 0, 0, 1]])

    # Camera Intrisecs matrix 3D -> 2D
    A2 = np.float32([[f, 0, w/2, 0],
                [0, f, h/2, 0],
                [0, 0, 1, 0]
        ])
    # # Translation matrix on the Z axis
    # T = np.float32([[[1], [0], [0], [0]],
    #     [[0], [1], [0], [0]],
    #     [[0], [0], [1], [dist]],
    #     [[0], [0], [0], [1]]])
    #
    # # Camera Intrisecs matrix 3D -> 2D
    # A2 = np.float32([[[f], [0], [w/2], [0]],
    #             [[0], [f], [h/2], [0]],
    #             [[0], [0], [1], [0]]
    #     ])

    print R.shape, A1.shape
    R_A = R.dot(A1)
    T_R_A = dot(T, R_A)
    H = dot(A2, T_R_A)
    
    print H.shape
#    H = A2 * (T * (R * A1));

    next_img_warp = cv2.warpPerspective(next_img_rgb, H.T, (img_w, img_h))

    # persp = cv.CreateImage(cv.GetSize(im), cv.IPL_DEPTH_8U, 3)
    # cv.WarpPerspective(im, persp, H2)
    #
    # H_inv = linalg.inv(m)
    # H_inv = np.identity(3)
    #
    # move_h = np.matrix(np.identity(3), np.float32)
    # move_h[0,2] = base_img_rgb.shape[1] * 0.25
    # move_h[1,2] = base_img_rgb.shape[0] * 0.25
    #
    # mod_inv_h = move_h * H_inv
    #
    # next_img_warp = cv2.warpPerspective(next_img_rgb, mod_inv_h, (img_w, img_h))

    fname, ext = os.path.splitext(os.path.basename(next_img))
    warp_file = os.path.join(warpedpath, fname+"_warp"+ext)
    cv2.imwrite(warp_file, next_img_warp)
    
# def saveWarpedImage(base_img, next_img, h):
#     global warpedpath
#
#     base_img_rgb = cv2.imread(base_img)
#     next_img_rgb = cv2.imread(next_img)
#
#     H = h
#     H = H / H[2,2]
#     H_inv = linalg.inv(H)
#
#     move_h = np.matrix(np.identity(3), np.float32)
#
#     img_w = int(base_img_rgb.shape[1] * 1.5)
#     img_h = int(base_img_rgb.shape[0] * 1.5)
#
#     move_h[0,2] = base_img_rgb.shape[1] * 0.25
#     move_h[1,2] = base_img_rgb.shape[0] * 0.25
#
#     mod_inv_h = move_h * H_inv
#
#     next_img_warp = cv2.warpPerspective(next_img_rgb, mod_inv_h, (img_w, img_h))
#
#     fname, ext = os.path.splitext(os.path.basename(next_img))
#     warp_file = os.path.join(warpedpath, fname+"_warp"+ext)
#     cv2.imwrite(warp_file, next_img_warp)
        
def readImageFilenames(dir_name):
    # read in image filenames
    dir_list = []
    for extension in (".png", ".jpg", ".jpeg"):
        try:
            temp = os.listdir(dir_name)
            temp = filter(lambda x: x.find(extension) > -1, temp)
            temp = sorted(temp)
            if len(temp) > 0:
                dir_list = temp
        except:
            print >> sys.stderr, ("Unable to open directory: %s" % dir_name)
            sys.exit(-1)

    dir_list = map(lambda x: os.path.join(dir_name,x), dir_list)

    #print len(dir_list), "files in directory"
    #print dir_list
    
    return dir_list
    
    
def readCamerasV2File(camerasfile):
    infile = open(camerasfile, 'r')
    
    lines = infile.readlines()
    print "\n"

    numcams = int(lines[16])
    
    camdict = {}
    
    camcenters = []
    xcents = []
    ycents = []

    sumx = 0
    sumy = 0
    sumz = 0
    
    R_accum = 0.
    
    for n in range(0,numcams):
        results = lines[18+n*14:31+n*14]

        newf = results[0].strip('\r\n')
        filename = results[1].strip('\r\n').split("\\")[-1]

        # Focal Length (of the undistorted image)
        focallength = (np.array(results[2].split(' ')))[0].astype(np.float)    

        # 2-vec Principal Point (image center)
        principalpoint = (np.array(results[3].split(' '))).astype(np.int)   

        # 3-vec Translation T (as in P = K[R T])
        T = (np.array(results[4].split(' '))).astype(np.float)
        
        # 3-vec Camera Position C (as in P = K[R -RC])
        C = (np.array(results[5].split(' '))).astype(np.float)  

        # 3-vec Axis Angle format of R
        R_axisangle = (np.array(results[6].split(' '))).astype(np.float)  

        R_accum += R_axisangle
        
        # 4-vec Quaternion format of R
        R_quat = (np.array(results[7].split(' '))).astype(np.float)  

        # 3x3 Matrix format of R
        R_mat = (np.array(" ".join(results[8:11]).split(' '))).astype(np.float)  
        R_mat = R_mat.reshape((3,3))
        
        # [Normalized radial distortion] = [radial distortion] * [focal length]^2
        normRadialDistortion = (np.array(results[11])).astype(np.float) 

        # # debug
        # print focallength, principalpoint
        # print T, C
        # print "R_axisangle: ",R_axisangle
        # print "R_quat:", R_quat
        # print "R_mat:", R_mat
        # print normRadialDistortion
        # print filename
        
        # print n, C[0]*10, C[2]*10
        # return results
        camdict[filename] = (focallength, T, C, R_mat, R_quat, normRadialDistortion, newf)
        
        xcents.append(C[0])
        ycents.append(C[2])
        
        sumx += C[0]
        sumy += C[2]
        sumz += C[1]
    
    # calculate averages over stack of images
    avgxcenter = sumx/numcams
    avgycenter = sumy/numcams
    avgzcenter = sumz/numcams

    avg_R = R_accum / numcams
    
    # plot to screen
    
    # xcents.append(avgxcenter)
    # ycents.append(avgycenter)
    # scatter(xcents, ycents)
    # show()

    xcents = np.array(xcents)
    ycents = np.array(ycents)
    
    print "========"
    print "average camera rotation:", avg_R
    print "range", (xcents.max()-xcents.min()), (ycents.max() - ycents.min())
    print "avg center:", avgxcenter, avgycenter, avgzcenter
    print "========"
    
    return camdict, np.array(avg_R).reshape((3,1)), (avgxcenter, avgycenter)
            
            
def newwarp(src, target, params1, params2):
    global warpedpath
    
    # image files
    src_img_rgb = cv2.imread(src)
    target_img_rgb = cv2.imread(target)
    
    w = src_img_rgb.shape[1]
    h = src_img_rgb.shape[0]
    print w,h
    
    # params for src image
    f1, T1, C1, R_mat1, R_quat1, normRadialDistortion1 = params1
    
    # params for target image
    f2, T2, C2, R_mat2, R_quat2, normRadialDistortion2 = params2
    
    # http://math.stackexchange.com/questions/87338/change-in-rotation-matrix
    
    print "R_mat1:", R_mat1
    # print "R_mat1T:", R_mat1.T
    
    print "R_mat2:", R_mat2
    
    # R_mat1 = R_mat1 / R_mat1[2,2]
    # R_mat2 = R_mat2 / R_mat2[2,2]
    #
    R_rod1, jacobian = cv2.Rodrigues(R_mat1)
    print "R_rod1:", R_rod1
    
    R_rod2, jacobian = cv2.Rodrigues(R_mat2)
    print "R_rod2:", R_rod2
    
    R_rod_12 = R_rod2 - R_rod1
    print "R_rod_12:", R_rod_12 * (180 / pi)
    
    R_12 = cv2.Rodrigues(R_rod_12)[0]
    print "R_12", R_12
    
    # # R_12 = R_12/R_12[2,2]
    # R_12_inv = linalg.inv(R_12)
    #
    # # H = np.identity(3) * T1.reshape((3,1))
    # H = R_mat1#R_12
    #
    # Tdiff = T2 - T1
    # print "Tdiff:", Tdiff

    # Cdiff = C1 - C2
    # print "Cdiff:", Cdiff
    
    # #Tdiff = Tdiff / Tdiff[2]
    # # print "Tdiff norm:", Tdiff
    #
    # M = np.identity(3)
    #
    # M[0,0] = f2
    # M[1,1] = f2
    #
    
    # H = np.identity(3)
    
    # H[0,2] =  Tdiff[0]
    # H[1,2] =  Tdiff[1]
    # H[2,2] =  Tdiff[2]

    # sf = 0.1
    # H[0,2] =  Cdiff[0] * w * sf
    # H[1,2] =  -1 * Cdiff[2] * h * sf
    # H[2,2] =  1

    # print "M:",M, "H:",H
    # H = M * H
    # print "M * H", M * H
    
    #
    # H[0,2] =  50
    # H[1,2] =  50
    # H[2,2] =  1
    
    # http://stackoverflow.com/questions/12288473/perspective-warping-in-opencv-based-on-know-camera-orientation
    
    x =  -14 * (pi/180)
    y =  20 * (pi/180)
    z =  15 * (pi/180)
    
    R_vec = np.array((x, y, z))
    print "R_vec:", R_vec
    R, jacobian = cv2.Rodrigues(R_vec)
    print "R:", R
    
    R = R_12 #R_mat1.T
    
    #Create trans mat and combine with translation matrix
    T = np.array([[1], [0], [-w/2],
                        [0], [1], [-h/2],
                        [0], [0], [1]]).reshape((3,3))
    # T[2,2] = 2
    # T = T / T[2,2]
    
    print "T:", T
    
    R[0,2] = 0
    R[1,2] = 0
    R[2,2] = 1
    
    trans = np.dot(R, T)
    trans[2,2] += w

    K = np.array([[f1], [0], [w/2],
                           [0], [f1], [h/2],
                           [0], [0], [1]]).reshape(3,3)
                        
    print "K:", K
    
    H = np.dot(K, trans)
    H = H / H[2,2]
    print "H", H
    
    src_img_warp = cv2.warpPerspective(src_img_rgb, H, (w, h))
    
    fname, ext = os.path.splitext(os.path.basename(src))
    warp_file = os.path.join(warpedpath, fname+"_warp"+ext)
    cv2.imwrite(warp_file, src_img_warp)

    # cv2.imshow("warped", src_img_warp)
    # cv2.waitKey()
    
def warpAvg(src, params, R_avg):
    global warpedpath

    # params for src image
    f1, T1, C1, R_mat1, R_quat1, normRadialDistortion1, undistort_fname = params

    # image files
    undistort = os.path.join("/Volumes/Work/Projects/lightfield/data/set2/undistorted/", undistort_fname)
    src_img_rgb = cv2.imread(undistort)

    w = src_img_rgb.shape[1]
    h = src_img_rgb.shape[0]


       # http://math.stackexchange.com/questions/87338/change-in-rotation-matrix

    print "R_mat1:", R_mat1
    # print "R_mat1T:", R_mat1.T

    # print "R_mat2:", R_mat2

    # R_mat1 = R_mat1 / R_mat1[2,2]
    # R_mat2 = R_mat2 / R_mat2[2,2]
    #
    R_rod1, jacobian = cv2.Rodrigues(R_mat1)
    print "R_rod1:", R_rod1

    print "R_avg:", R_avg
    
    # R_rod2, jacobian = cv2.Rodrigues(R_mat2)
    # print "R_rod2:", R_rod2

    R_rod_12 = R_avg - R_rod1 #R_rod2 - R_rod1
    print "R_rod_12:", R_rod_12 * (180 / pi)

    R_12, jacobian = cv2.Rodrigues(R_rod_12)
    print "R_12:", R_12

    # # R_12 = R_12/R_12[2,2]
    # R_12_inv = linalg.inv(R_12)
    #
    # # H = np.identity(3) * T1.reshape((3,1))
    # H = R_mat1#R_12
    #
    # Tdiff = T2 - T1
    # print "Tdiff:", Tdiff

    # Cdiff = C1 - C2
    # print "Cdiff:", Cdiff

    # #Tdiff = Tdiff / Tdiff[2]
    # # print "Tdiff norm:", Tdiff
    #
    # M = np.identity(3)
    #
    # M[0,0] = f2
    # M[1,1] = f2
    #

    # H = np.identity(3)

    # H[0,2] =  Tdiff[0]
    # H[1,2] =  Tdiff[1]
    # H[2,2] =  Tdiff[2]

    # sf = 0.1
    # H[0,2] =  Cdiff[0] * w * sf
    # H[1,2] =  -1 * Cdiff[2] * h * sf
    # H[2,2] =  1

    # print "M:",M, "H:",H
    # H = M * H
    # print "M * H", M * H

    #
    # H[0,2] =  50
    # H[1,2] =  50
    # H[2,2] =  1

    # http://stackoverflow.com/questions/12288473/perspective-warping-in-opencv-based-on-know-camera-orientation

    R = R_12 #R_mat1.T

    #Create trans mat and combine with translation matrix
    T = np.array([[1], [0], [-w/2],
                       [0], [1], [-h/2],
                       [0], [0], [1]]).reshape((3,3))
    # T[2,2] = 2
    # T = T / T[2,2]

    print "T:", T

    R[0,2] = 0
    R[1,2] = 0
    R[2,2] = 1

    trans = np.dot(R, T)
    trans[2,2] += w

    K = np.array([[f1*0.8], [0], [w/2],
                  [0], [f1*0.8], [h/2],
                  [0], [0], [1]]).reshape(3,3)
                
    print "K:", K

    H = np.dot(K, trans)
    H = H / H[2,2]
    print "H", H

    src_img_warp = cv2.warpPerspective(src_img_rgb, H, (w, h))

    fname, ext = os.path.splitext(os.path.basename(src))
    warp_file = os.path.join(warpedpath, fname+"_warp"+ext)
    cv2.imwrite(warp_file, src_img_warp)

               
               
if __name__ == '__main__':
    global featurepath, warpedpath

    # Define command line argument interface
    parser = argparse.ArgumentParser(description='apply Visual SFM quaternions to a folder of lightfield images')
    parser.add_argument('inpath', help='path to input images')
    parser.add_argument('nvmfile', help='N-view match file')
    parser.add_argument('warpedpath', help='path to warped output image files')
    
    args = parser.parse_args()

    dir_name = args.inpath  # directory of input images
    nvmfile = args.nvmfile  # results of N-view match from Visual SFM in windows
    warpedpath = args.warpedpath    # save the results
    
    if not os.path.exists(warpedpath):
        os.makedirs(warpedpath)
    
    # read image filenames
    image_files = readImageFilenames(dir_name)
    
    # read Cameras V2 file    
    camresults, R_avg, C_avg = readCamerasV2File(nvmfile)
    
    firstimg = os.path.basename(image_files[0])
    #print firstimg
    #print camresults
    #print camresults[firstimg]
    
    # warp image according to nvm results
    
    for src in image_files:
        print "source:", src
        src_fname = os.path.basename(src)
        warpAvg(src, camresults[src_fname], R_avg)

    # count = 0
    # for src in image_files:
    #     src_fname = os.path.basename(src)
    #     C_src = camresults[src_fname][2]
    #     print count, (C_avg[0]-C_src[0]), (C_avg[1]-C_src[2])
    #     count = count + 1


    # for src in image_files:
    #     print "target:", target
    #     print "source:", src
    #     src_fname = os.path.basename(src)
    #     newwarp(src, target, camresults[src_fname], camresults[target_fname])
    #
    # count = 0
    # for src in image_files:
    #     # print "target:", target
    #     # print "source:", src
    #     src_fname = os.path.basename(src)
    #     C_src = camresults[src_fname][2]
    #     C_target = camresults[target_fname][2]
    #     print count, (C_target[0]-C_src[0]), (C_target[2]-C_src[2])
    #     count = count + 1
    #
